# Copyright 2019-2020 LiveRamp Holdings, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

require 'countries'

require 'terraorg/model/people'
require 'terraorg/model/util'

class Squad
  attr_accessor :id, :name, :metadata, :teams

  class Team
    attr_accessor :location, :members, :associates

    def initialize(parsed_data, people)
      location = parsed_data.fetch('location')
      country = ISO3166::Country.new(location)
      raise "Location is invalid: #{location}" unless country
      @location = country.alpha2
      @members = parsed_data.fetch('members', []).map do |n|
        people.get_or_create!(n)
      end
      @associates = parsed_data.fetch('associates', []).map do |n|
        people.get_or_create!(n)
      end
    end

    def validate!
      raise 'Subteam has no full time members' if @members.size == 0
      # location validation done at initialize time
      # associates can be empty

      # associates and members must have zero intersection
      associate_set = Set.new(@associates.map(&:id))
      member_set = Set.new(@members.map(&:id))
      raise 'A member cannot also be an associate of the same team' if associate_set.intersection(member_set)
    end

    # Output a canonical (sorted, formatted) version of this Team.
    # - Sort the members in each team
    # - Only add an associates field if it's present
    def to_h
      {
        'associates' => @associates.map(&:id).sort,
        'location' => @location,
        'members' => @members.map(&:id).sort,
      }
    end

    def everyone
      @associates + @members
    end

    def to_md
      "**#{@location}**: #{@members.map(&:name).sort.join(', ')}, #{@associates.map { |m| "_#{m.name}_" }.sort.join(', ')}"
    end
  end

  def initialize(id, parsed_data, people, gsuite_domain, slack_domain)
    @gsuite_domain = gsuite_domain
    @slack_domain = slack_domain
    @id = id
    @metadata = parsed_data.fetch('metadata', {})
    @name = parsed_data.fetch('name')
    @people = people

    teams_arr = parsed_data.fetch('team', []).map do |t|
      Team.new(t, people)
    end
    @teams = Hash[teams_arr.map { |t| [t.location, t] }]
  end

  # Everyone including associates on all subteams in the squad.
  def everyone(location: nil)
    @teams.select { |l, t|
      location == nil || l == location
    }.map { |_, t|
      t.everyone
    }.flatten
  end

  # Full-time members of all subteams in this squad
  def members
    @teams.map { |_, t|
      t.members
    }.flatten
  end

  def get_acl_groups(org_id)
    # each geographically located subteam
    groups = Hash[@teams.map { |location, team|
      [unique_name(org_id, location), {'name' => "#{@name} squad members based in #{location}", 'members' => team.everyone}]
    }]

    # combination of all subteams
    groups[unique_name(org_id, nil)] = {'name' => "#{@name} squad worldwide members", 'members' => everyone}

    groups
  end

  def unique_name(org_id, location)
    if location
      "#{org_id}-squad-#{@id}-#{location.downcase}"
    else
      "#{org_id}-squad-#{@id}"
    end
  end

  def validate!
    @teams.each(&:validate!)
  end

  def to_md(platoon_name, org_id)
    pm = @metadata.fetch('pm', [])
    pm = pm.map { |p| @people.get_or_create!(p).name }.join(', ')

    sme = @metadata.fetch('sme', '')
    if !sme.empty?
      sme = @people.get_or_create!(sme).name
    end

    manager = @metadata.fetch('manager', '')
    if !manager.empty?
      manager = @people.get_or_create!(manager).name
    end

    subteam_members = @teams.values.map(&:to_md).join(' / ')
    email = "#{unique_name(org_id, nil)}@#{@gsuite_domain}"
    slack = @metadata.fetch('slack', '')
    if slack
      slack = "[#{slack}](https://#{@slack_domain}/app_redirect?channel=#{slack.gsub(/^#/, '')})"
    end
    # platoon name, squad name, PM, email list, SME, slack, # full time members, squad manager, members
    "|#{platoon_name}|#{@name}|#{pm}|[#{email}](#{email})|#{sme}|#{slack}|#{members.size}|#{manager}|#{subteam_members}|"
  end

  def generate_tf(org_id)
    groups = get_acl_groups(org_id)

    groups.map { |id, group|
      description = "#{group.fetch('name')} (terraorg)"
      <<-EOF
resource "okta_group" "#{id}" {
  name = "#{id}"
  description = "#{description}"
  users = #{Util.persons_tf(group.fetch('members'))}
}

#{Util.gsuite_group_tf(id, @gsuite_domain, group.fetch('members'), description)}
EOF
    }.join("\n\n")
  end

  def manager
    m = @metadata['manager']
    if m
      return @people.get_or_create!(m)
    else
      return nil
    end
  end

  def to_h
    # Output a canonical (sorted, formatted) version of this Squad.
    # - Subteams are sorted by location lexically
    obj = { 'id' => @id, 'name' => @name }
    obj['team'] = @teams.values.sort_by { |t| t.location }.map(&:to_h)
    obj['metadata'] = @metadata if @metadata

    obj
  end
end
